(* Implementing a FIFO *)

module type FIFO = sig
  type 'a t
  exception Empty
  val empty : 'a t
  val is_empty : 'a t -> bool
  val add : 'a -> 'a t -> 'a t  
  val remove : 'a t -> ('a * 'a t)  
  (** raise exception [Empty] *)
end;;

module FifoNaive : FIFO = struct
  type 'a t = 'a list
  exception Empty
  let empty = []
  let is_empty f = f = []
  let add a f = f@[a]
  let remove = function
  | [] -> raise Empty
  | a::l -> (a, l)
end;;

module type FIFOIMP = sig
  type 'a t
  exception Empty
  val create : unit -> 'a t
  val is_empty : 'a t -> bool
  val add : 'a -> 'a t -> unit
  val remove : 'a t -> 'a
  (** raises [Empty] if the queue is empty *)
end;;

module FifoImp = struct
  exception Empty

  type 'a cell = 
         Vide 
       | Cons of 'a * 'a cell ref
  type 'a t = {mutable first: 'a cell; 
               mutable last : 'a cell}

  let create () = {first = Vide; last = Vide}

  let is_empty f = f.first = Vide

  let add a f = match f.last with
    | Vide -> (* f.first doit etre Vide aussi *)
      f.first <- Cons (a, ref Vide);
                 f.last <- f.first;
    | Cons (_, r) -> r := Cons (a, ref Vide); 
                     f.last <- !r

  let remove f = match f.first with
  | Vide -> raise Empty
  | Cons (a, r) -> 
        if f.last = f.first then f.last <- !r;
        f.first <- !r ; a
end;;


(* The efficient solution *)

module FifoDL : FIFO = struct
  exception Empty

  type 'a t = 'a list * 'a list

  let empty = ([], [])

  let is_empty = 
  function
    ([],[]) -> true 
  |   _     -> false

  let add x (l1, l2) = (x::l1, l2)

  let remove (l1, l2) = match l2 with
  | a::l -> (a, (l1, l))
  | [] -> match List.rev l1 with
      | [] -> raise Empty
      | a::l -> (a, ([], l))
end;;


(** Binary search trees *)

type 'a abr = E
            | T of 'a abr * 'a * 'a abr
;;

(* Red-Black trees *)

type color = R | B
 type tree = E | T of color * tree * elem * tree;;

let rec member x = function
  | E -> false
  | T (_, a, y, b) ->
      if islt x y then member x a
      else if islt y x then member x b
      else true;;

let bal = function
  | B, T (R, T (R, a, x, b), y, c), z, d
  | B, T (R, a, x, T (R, b, y, c)), z, d
  | B, a, x, T (R, T (R, b, y, c), z, d)
  | B, a, x, T (R, b, y, T (R, c, z, d)) ->
      T (R, T (B, a, x, b), y, T (B, c, z, d))
  | a, b, c, d -> T (a, b, c, d);;

let insert x s =
  let rec ins = function
    | E -> T (R, E, x, E)
    | T (color, a, y, b) as s ->
        if islt x y then bal (color, ins a, y, b)
        else if islt y x then 
                         bal (color, a, y, ins b)
        else s in
  match ins s with  (* surement non vide *)
  | T (_, a, y, b) -> T (B, a, y, b)
  | _ -> assert false;;


(* Using Red-Black trees to build a Set module *)

module type ORDERED = sig
  type t
  val compare : t -> t -> int
end

module type SET = sig
  type elem
  type set
  val empty : set
  val insert : elem -> set -> set
  val member : elem -> set -> bool
end;;

module RedBlackSet (Element : ORDERED) : 
      (SET with type elem = Element.t) = 
struct

 type elem = Element.t
 let islt x y = (Element.compare x y) < 0

 type color = R | B
 type tree = E | T of color * tree * elem * tree
 type set = tree

 let empty = E

 let rec member x = function
   | E -> false
   | T (_, a, y, b) ->
       if islt x y then member x a
       else if islt y x then member x b
       else true

 let bal = function
   | B, T (R, T (R, a, x, b), y, c), z, d
   | B, T (R, a, x, T (R, b, y, c)), z, d
   | B, a, x, T (R, T (R, b, y, c), z, d)
   | B, a, x, T (R, b, y, T (R, c, z, d)) ->
       T (R, T (B, a, x, b), y, T (B, c, z, d))
   | a, b, c, d -> T (a, b, c, d)

 let insert x s =
   let rec ins = function
     | E -> T (R, E, x, E)
     | T (color, a, y, b) as s ->
         if islt x y then bal (color, ins a, y, b)
         else if islt y x 
              then bal (color, a, y, ins b)
              else s in
   match ins s with  (* surement non vide *)
   | T (_, a, y, b) -> T (B, a, y, b)
   | _ -> assert false
end
;;
